# coding=utf-8
"""Implements the table field."""
import typing

from . import api, exceptions, detail


_NO_DEFAULT = object()


_TableField = typing.NamedTuple(
  "_TableField",
  (
    ("header", str),
    ("required", bool),
    ("default", typing.Optional[object]),
    ("field_formatter", api.FieldFormatter),
    ("field_empty_check", api.FieldEmptyCheck),
    ("field_reader", api.FieldReader),
  )
)


class DefaultField(_TableField):
  """Default implementation for api.Field."""

  def __call__(self, row):
    """Extract value from row, format it, handling defaults."""

    try:
      cell_value = self.field_reader(field=self, instance=row)
    except exceptions.MissingField:
      cell_value = None

    if self.field_empty_check(field=self, instance=cell_value):
      if self.required and self.default is _NO_DEFAULT:
        raise exceptions.MissingField(self, row)
      if self.default is not _NO_DEFAULT:
        cell_value = self.default
    if self.field_formatter:
      cell_value = self.field_formatter(field=self, instance=cell_value)
    return cell_value


class FieldFactory(object):
  """Implements a field factory."""

  @classmethod
  def create_const_field(cls, header: str, value: object) -> detail.TableField:
    """
    Create field that returns value irregardless of data what is in the row.
    """
    return cls(header=header, required=False).const_field(value).noop_empty_check().create()

  def __init__(
      self,
      header: str,
      required: bool = False,
      default: object = _NO_DEFAULT
  ):
    self.header = header
    self.required = required
    self.default = default
    self.field_reader = None
    self.field_formatter = detail.NoopFormatter()
    self.field_empty_check = detail.DefaultFieldEmptyCheck()

  def path(self, path: str) -> 'FieldFactory':
    """Read cell value using a path."""
    self.field_reader = detail.DefaultFieldReader(path=path)
    return self

  def const_field(self, const: object) -> 'FieldFactory':
    """
    Always return constant value for this field.

    This is useful if column is just missing from source data, yet must be present in output data.
    """
    self.field_reader = detail.ConstReader(instance=const)
    return self

  def reader(self, field_reader: 'api.FieldReader'):
    """
    Directly set field reader.
    """
    self.field_reader = field_reader
    return self

  def formatter(self, formatter: 'api.FieldFormatter') -> 'FieldFactory':
    """Set formatter explicitly."""
    self.field_formatter = formatter
    return self

  def type_formatter(self, object_type: type) -> 'FieldFactory':
    """
    Use formatter that coalesces to given built-in type.

    For example to format values as integers use: "self.type_formatter(int)".
    """
    self.field_formatter = detail.TypeFormatter(object_type=object_type)
    return self

  def format_formatter(self, fformat: str, object_type: type = None):
    """Adds formatter that uses str.format."""
    self.field_formatter = detail.FormatFormatter(fformat, object_type)
    return self

  def empty_check(self, empty_check: 'api.FieldEmptyCheck'):
    """Set empty-check explicitly."""
    self.field_empty_check = empty_check
    return self

  def enumerated_empty_check(self, empty_values: typing.Sequence):
    """
    Use empty check that treats special values as empty.

    For example use enumerated_empty_check([None, -1]) to treat None and -1 as missing values.
    """
    self.field_empty_check = detail.EnumeratedFieldEmptyCheck(empty_values=empty_values)
    return self

  def noop_empty_check(self):
    """Disable empty checking. """
    self.field_empty_check = detail.NoopEmptyCheck()
    return self

  def create(self) -> DefaultField:
    """Create the field."""
    return DefaultField(
      header=self.header,
      field_reader=self.field_reader,
      field_empty_check=self.field_empty_check,
      required=self.required,
      default=self.default,
      field_formatter=self.field_formatter,
    )



# coding=utf-8
# pylint: disable=missing-docstring


import pytest

from xls_writer import detail
from xls_writer.detail import FormatFormatter

noop_formater_test_data = list(range(10)) + [None, "foo", "bar"]


@pytest.mark.parametrize("value", noop_formater_test_data)
def test_default_formatter(value):
  noop = detail.NoopFormatter()
  assert noop(field=None, instance=value) is value


type_formatter_test_data = (
  (int, "1", 1),
  (int, 1, 1),
  (int, 1.0, 1),
  (float, 1, 1.0),
  (str, 1, "1")
)


@pytest.mark.parametrize("object_type, value, expected", type_formatter_test_data)
def test_default_empty_check(object_type, value, expected):
  check = detail.TypeFormatter(object_type=object_type)
  assert expected == check(field=None, instance=value)


@pytest.mark.parametrize("field_format, obj_type, instance, expected", (
  ("{:d}", None, 1, "1"),
  ('{:d}', int, 1, "1"),
  ('{:06.2f}', None, 3.141, "003.14"),
  ('{:06.2f}', float, 3.141, "003.14"),
  ('{}', int, 3.141, "3"),
  ('{foo}: {bar}', None, {'foo': 1, 'bar': 2}, '1: 2')
))
def test_format_formatter(field_format, obj_type, instance, expected):
  formatter = FormatFormatter(field_format, obj_type)
  assert expected == formatter(field=None, instance=instance)



# coding=utf-8
# pylint: disable=missing-docstring,redefined-outer-name,blacklisted-name,invalid-name

import csv
import io

import pytest

from xls_writer import api, utils, csv_formatter
from xls_writer.field import FieldFactory


@pytest.fixture()
def source_data():

  class Bar(object):
    def __init__(self, canary):
      self.baz = canary

  return [
    {"foo": "1", "bar": Bar("row 1")},
    {"foo": "2", "bar": Bar("row 2")},
    {"foo": 3}
  ]


@pytest.fixture()
def table_description():
  fields = [
    FieldFactory("Foo Header", required=True).path("foo").type_formatter(int).create(),
    FieldFactory("Bar Header", required=False, default="<<Missing>>")
    .path("bar.baz").type_formatter(str).create(),
    FieldFactory.create_const_field("Description", "Refer to terms and conditions"),
    FieldFactory.create_const_field("Custom Field 123", "")
  ]
  return api.TableDescription(fields=fields)


expected_result = (
  "Foo Header\tBar Header\tDescription\tCustom Field 123\r\n"
  "1\trow 1\tRefer to terms and conditions\t\r\n"
  "2\trow 2\tRefer to terms and conditions\t\r\n"
  "3\t<<Missing>>\tRefer to terms and conditions\t\r\n"
)


def test_csv(source_data, table_description):
  file = io.StringIO()
  formatter = csv_formatter.CSVFormatterFactory(dialect=csv.excel_tab)
  utils.store_table(source_data, table_description, formatter, file, close_file_on_finish=False)
  file.seek(0)
  actual = str(file.read(-1))
  assert actual == expected_result


@pytest.mark.parametrize('close', [True, False])
def test_formatter_closes_file_when_asked(close):
  file = io.StringIO()
  formatter = csv_formatter.CSVFormatterFactory(dialect=csv.excel_tab).create()
  formatter.initialize_file(file_object=file)
  formatter.close(close_underlying_file=close)
  assert file.closed == close



# coding=utf-8
"""CSV Formatter"""

import csv
import typing

from . import api


class CSVFormatter(api.TableFormatter):
  """CSV formatter"""

  def __init__(self, dialect: csv.Dialect=csv.excel_tab):
    super().__init__()
    self.dialect = dialect
    self.writer = None

  def initialize_file(self, file_object):
    super().initialize_file(file_object)
    self.writer = csv.writer(self.file_object, dialect=self.dialect)

  def write_row(self, row: typing.Sequence):
    self.writer.writerow(row)

  @property
  def type(self) -> api.TableFormatterType:
    return api.TableFormatterType.STRING


class CSVFormatterFactory(api.FormatterFactory):
  """CSV Formatter factory"""

  def __init__(self, dialect: csv.Dialect=csv.excel_tab):
    super().__init__()
    self.dialect = dialect

  def create(self) -> api.TableFormatter:
    return CSVFormatter(dialect=self.dialect)




# coding=utf-8
"""
Implementation details. These classes should be transparent to end user.
"""
import typing

from xls_writer.api import TableField
from . import exceptions, api


# pylint: disable=no-self-use


class EnumeratedFieldEmptyCheck(api.FieldEmptyCheck):
  """
  Checks if field is empty by comparing it to set of known empty values.
  Useful if you have dumb API that, for example, treats -1 as none.
  """

  def __init__(self, empty_values: typing.Sequence):
    self.empty_values = set(empty_values)

  def __call__(self, field: api.TableField, instance) -> bool:
    return instance in self.empty_values


class DefaultFieldEmptyCheck(api.FieldEmptyCheck):
  """
  Field empty check that treats only Nones as empty values.
  """

  def __call__(self, field: api.TableField, instance) -> bool:
    return instance is None


class NoopEmptyCheck(api.FieldEmptyCheck):
  """
  Values are never treated as empty.
  """

  def __call__(self, field: TableField, instance) -> bool:
    return False


class TypeFormatter(api.FieldFormatter):
  """
  Formatter that coalesces to given type. That is: calls ``int(value)``.
  """
  def __init__(self, object_type: type):
    """
    :param object_type: Type to coalesce to.
    """
    self.object_type = object_type

  def __call__(self, field: TableField, instance) -> object:
    return self.object_type(instance)


class FormatFormatter(api.FieldFormatter):
  """
  Formatter that uses str.format.
  """

  def __init__(self, field_format: str, object_type: type = None):
    super().__init__()
    self.format = field_format
    self.type = object_type

  def __call__(self, field: TableField, instance) -> object:
    if self.type is not None:
      instance = self.type(instance)
    if isinstance(instance, dict):
      return self.format.format(**instance)
    return self.format.format(instance)


class NoopFormatter(api.FieldFormatter):
  """
  Formatter that does nothing.
  """

  def __call__(self, field: TableField, instance) -> object:
    return instance


class ConstReader(api.FieldReader):
  """
  Reader that always returns the same value.
  """

  def __init__(self, instance):
    self.instance = instance

  def __call__(self, field: TableField, instance) -> object:
    return self.instance


class MultiReader(api.FieldReader):
  """Reader that proxies reading to multiple fields. It returns a dict."""

  def __init__(self, child_fields: typing.Mapping[str, api.TableField]):
    self.child_fields = child_fields

  def __call__(self, field: TableField, instance) -> object:
    return {
      field_key: child_field(row=instance)
      for field_key, child_field in self.child_fields.items()
    }


class DefaultFieldReader(api.FieldReader):
  """
  Field reader that extracts row values by traversing object tree using given path.
  """

  def __init__(self, path: str):
    """
    :param path: Dotted path to field.
    """
    self.path = path.split(".")

  def __get_field(self, inst, fieldname):
    _does_not_exist = object()
    value = getattr(inst, fieldname, _does_not_exist)
    if value is not _does_not_exist:
      return value
    try:
      return inst[fieldname]
    except (KeyError, IndentationError, TypeError):
      pass
    raise exceptions.MissingField()

  def __call__(self, field: TableField, instance) -> object:
    result = instance
    for path_item in self.path:
      try:
        result = self.__get_field(result, path_item)
      except exceptions.MissingField:
        raise exceptions.MissingField(self.path, path_item, instance)
    return result



# coding=utf-8
"""Module with API classes."""
import abc
import enum
import typing


class TableField(object):
  """Represents table field."""

  @property
  def header(self) -> str:
    """Field header"""
    raise NotImplementedError

  def __call__(self, row: object) -> object:
    """
    Extracts and formats data from row.
    """


class BaseToString(object):
  """Base class that gives meaningful __repr__"""

  def __repr__(self):  # pragma: no cover
    return "<{typename} {dict}>".format(
      typename=type(self).__name__,
      dict=" ".join(["{0}='{1}'".format(*x) for x in sorted(self.__dict__.items())])
    )


class FieldReader(BaseToString, metaclass=abc.ABCMeta):
  """Object that extracts data from row."""

  @abc.abstractmethod
  def __call__(self, field: TableField, instance) -> object:
    """
    Extract data from row.

    :param field: Field this reader belongs to.
    :param instance: Row instance.
    :return: Cell value
    """
    raise NotImplementedError


class FieldFormatter(BaseToString, metaclass=abc.ABCMeta):
  """Formats cell value."""

  @abc.abstractmethod
  def __call__(self, field: TableField, instance) -> object:
    """

    :param field: Field this formatter belongs to.
    :param instance: Cell value.
    :return: formatted value
    """
    raise NotImplementedError


class FieldEmptyCheck(BaseToString, metaclass=abc.ABCMeta):
  """
  Checks if field value is empty. If field is empty it will be replaced with default value
  """
  @abc.abstractmethod
  def __call__(self, field: TableField, instance) -> bool:
    """

    :param field: Field this object belongs to.
    :param instance: Extracted cell value (before formatting!)
    :return: true if field IS EMPTY.
    """
    raise NotImplementedError


class TableFormatterType(enum.Enum):
  """Type of formatter: it either formats table to byte stream or string stream."""
  STRING = "STRING"
  BYTES = "BYTES"


class TableFormatter(object, metaclass=abc.ABCMeta):
  """Object that formats table to either byte stream or char stream."""

  def __init__(self):
    super().__init__()
    self.file_object = None

  def initialize_file(self, file_object):
    """
    :param file_object: File opened either in binary mode (if type is BYTES or in char mode)
    """
    self.file_object = file_object

  @property
  @abc.abstractmethod
  def type(self) -> TableFormatterType:
    """
    Returns whether this formatter writes rows as bytes or
    """
    raise NotImplementedError

  def write_header(self, header: typing.Sequence[str]):
    """
    Should be called before write row.
    :param header: Header to be written
    """
    self.write_row(header)

  @abc.abstractmethod
  def write_row(self, row: typing.Sequence):
    """
    :param row: Row to be written
    """
    raise NotImplementedError

  def close(self, close_underlying_file: bool):
    """
    Closes underlying file.
    """
    self.file_object.flush()
    if close_underlying_file:
      self.file_object.close()


class FormatterFactory(object, metaclass=abc.ABCMeta):
  """Factory for formatters."""

  @abc.abstractmethod
  def create(self) -> TableFormatter:
    """Creates the formatter."""
    raise NotImplementedError


class TableDescription(object):
  """Immutable object representing a table."""

  def __init__(self, fields=typing.Sequence[TableField]):
    self.fields = tuple(fields)

  @property
  def header(self) -> typing.List[str]:
    """Table header."""
    return [
      field.header
      for field in self.fields
    ]

  def extract_row(self, row: object) -> typing.List[object]:
    """Table row."""
    return [
      field(row)
      for field in self.fields
    ]



# coding=utf-8

# pylint: disable=missing-docstring,blacklisted-name


def foobarbaz_dict(canary="canary"):
  return {"foo": {"bar": {"baz": canary}}}


def foobarbaz_object(canary="canary"):
  class Baz(object):
    def __init__(self):
      self.baz = canary

  class BarBaz(object):
    def __init__(self):
      self.bar = Baz()

  class FooBarBaz(object):
    def __init__(self):
      self.foo = BarBaz()

  return FooBarBaz()


def foobarbaz_mixed(canary="canary"):
  class Baz(object):
    def __init__(self):
      self.baz = canary

  return {"foo": {"bar": Baz()}}


def foobarbaz(canary="canary"):
  return [foobarbaz_dict(canary), foobarbaz_mixed(canary), foobarbaz_object(canary)]
